import os

NAME = "abcli"

ICON = "🪄"

DESCRIPTION = f"{ICON} a language to speak AI."

VERSION = "9.310.1"

REPO_NAME = "awesome-bash-cli"

MARQUEE = "https://github.com/kamangir/awesome-bash-cli/raw/main/assets/marquee.png"



from functools import reduce
import os

from blue_objects import file

from abcli import env
from abcli.table import Table
from abcli.logger import logger


_, inverse_of = file.load_json(
    os.path.join(
        env.abcli_path_abcli,
        "abcli/plugins/relations/relations.json",
    )
)
inverse_of.update({inverse_of[relation]: relation for relation in inverse_of})
inverse_of[""] = ""

list_of = sorted(list(set(list(inverse_of.keys()) + list(inverse_of.values()))))


columns = "object_1,object_2,relation,timestamp".split(",")


def clone(object_1, object_2):
    """clone object_1 relations -> object_2.

    Args:
        object_1 (str): object 1.
        object_2 (str): object 2.

    Returns:
        bool: success
    """
    return reduce(
        lambda x, y: x and y,
        [
            reduce(
                lambda x, y: x and y,
                [set_(object_2, object, relation) for object in object_list],
                True,
            )
            for relation, object_list in search(object_1).items()
        ],
        True,
    )


def create():
    """create the relations table.

    Returns:
        bool: success.
    """
    return Table.Create(
        "relations",
        [
            "object_1 VARCHAR(256) NOT NULL",
            "object_2 VARCHAR(256) NOT NULL",
            "relation VARCHAR(256) NOT NULL",
        ],
    )


def get(object_1, object_2):
    """get relation between object_1 and object_2.

    Args:
        object_1 (str): object 1.
        object_2 (str): object 2.

    Returns:
        str: relation.
    """

    if object_1 > object_2:
        return inverse_of[get(object_1=object_2, object_2=object_1)]

    table = Table(name="relations")

    if not table.connect():
        return ""

    success, output = table.execute(
        "SELECT r.relation "
        "FROM abcli.relations r "
        "INNER JOIN ( "
        "SELECT MAX(timestamp) AS max_timestamp "
        "FROM abcli.relations "
        f'WHERE object_1="{object_1}" AND object_2="{object_2}" '
        ") rm "
        "ON r.timestamp=rm.max_timestamp "
        f'WHERE object_1="{object_1}" AND object_2="{object_2}";'
    )
    if not success:
        return ""

    if not table.disconnect():
        return ""

    return "" if not output else output[-1][0]


def search(object, relation="", count=-1):
    """search for relations.

    Args:
        object (str): object
        relation (str, optional): only this relation. Defaults to "".
        count (int, optional): return count, only relevant if relation. Defaults to -1 (all).

    Returns:
        Any: {"relation_1": ["object_1","object_2"], "relation_2": ["object_3" ,"object_4"]} or
        ["object_1", "object_2"] if relation is given.
    """
    table = Table(name="relations")

    output = [] if relation else {}

    if not table.connect():
        return output

    success, output_right = table.execute(
        "SELECT r.relation, r.object_2 "
        "FROM abcli.relations r "
        "INNER JOIN (  "
        "SELECT object_2, MAX(timestamp) AS max_timestamp "
        "FROM abcli.relations "
        f'WHERE object_1="{object}" GROUP BY object_2'
        ") rm  "
        "ON r.timestamp=rm.max_timestamp "
        f'WHERE object_1="{object}"; '
    )
    if not success:
        return {}

    success, output_left = table.execute(
        "SELECT r.relation, r.object_1 "
        "FROM abcli.relations r "
        "INNER JOIN (  "
        "SELECT object_1, MAX(timestamp) AS max_timestamp "
        "FROM abcli.relations "
        f'WHERE object_2="{object}" GROUP BY object_1'
        ") rm  "
        "ON r.timestamp=rm.max_timestamp "
        f'WHERE object_2="{object}"; '
    )
    if not success:
        return output

    if not table.disconnect():
        return output

    raw_output = {thing[1]: thing[0] for thing in output_right}
    raw_output.update({thing[1]: inverse_of[thing[0]] for thing in output_left})

    raw_output = {thing: relation for thing, relation in raw_output.items() if relation}

    output = {}
    for object_, relation_ in raw_output.items():
        output[relation_] = output.get(relation_, []) + [object_]

    if not relation:
        return output

    output = output.get(relation, [])

    if count != -1:
        output = output[-count:]

    return output


def set_(object_1, object_2, relation):
    """set relation between object_1 and object_2.

    Args:
        object_1 (str): object 1.
        object_2 (str): object 2.
        relation (str): relation.

    Returns:
        bool: success.
    """
    if relation not in list_of:
        logger.error(f"-relations: set: {relation}: relation not found.")
        return False

    if object_1 > object_2:
        return set_(object_2, object_1, inverse_of[relation])

    table = Table(name="relations")

    if not table.connect():
        return False

    if not table.insert(
        ["object_1", "object_2", "relation"],
        [object_1, object_2, relation],
    ):
        return False

    if not table.disconnect():
        return False

    logger.info("{} ={}=> {}".format(object_1, relation if relation else "X", object_2))

    return True



from functools import reduce
import random
import re
from blue_options.options import Options
from abcli.table import Table
from abcli.logger import logger


def clone(object_1, object_2):
    """clone object_1 tags -> object_2.

    Args:
        object_1 (str): object 1.
        object_2 (str): object 2.

    Returns:
        bool: success
    """
    return set_(object_2, get(object_1))


def create():
    """create the tags table.

    Returns:
        bool: success.
    """
    return Table.Create(
        "tags",
        [
            "keyword VARCHAR(256) NOT NULL",
            "tag VARCHAR(4096) NOT NULL",
            "value BIT NOT NULL",
        ],
    )


def get(keyword):
    """get tags for keyword.

    Args:
        keyword (str): keyword.

    Returns:
        bool: List(str).
    """
    table = Table(name="tags")

    if not table.connect():
        return []

    success, output = table.execute(
        "SELECT t.tag,t.value "
        f"FROM {table.name} t "
        "INNER JOIN ( "
        "SELECT tag, MAX(timestamp) AS max_timestamp "
        f"FROM {table.name} "
        f'WHERE keyword="{keyword}" GROUP BY tag '
        ") tm "
        "ON t.tag=tm.tag AND t.timestamp=tm.max_timestamp "
        f'WHERE keyword="{keyword}";',
    )

    if success:
        success = table.disconnect()

    if not success:
        return []

    return sorted([thing[0] for thing in output if thing[1] == b"\x01"])


def search(
    tags,
    after="",
    before="",
    count=-1,
    host=-1,
    return_timestamp=False,
    shuffle=False,
    offset=0,
):
    """search.

    Args:
        tags (List(str)): list of tags.
        after (str, optional): include keywords after and including. Defaults to "".
        before (str, optional): include keywords before and including. Defaults to "".
        count (int, optional): keyword count to return. Defaults to -1.
        host (int, optional): limit to/exclude/ignore (1/0/-1) hosts. Defaults to -1.
        return_timestamp (bool, optional): return timestamp. Defaults to False.
        shuffle (bool, optional): shuffle output. Defaults to False.
        offset (int, optional): offset. Defaults to 0.

    Returns:
        List(str): list of keywords.
    """
    if isinstance(tags, str):
        tags = tags.split(",")

    included_tags = []
    excluded_tags = []
    for tag in tags:
        if tag:
            if tag[0] in "~-!":
                excluded_tags += [tag[1:]]
            else:
                included_tags += [tag]

    table = Table(name="tags")

    table.connect()

    list_of_keywords = None
    timestamp = {}
    for tag in included_tags:
        success, output = table.execute(
            "SELECT t.keyword,t.value,t.timestamp "
            "FROM abcli.tags t "
            "INNER JOIN ( "
            "SELECT keyword, MAX(timestamp) AS max_timestamp "
            "FROM abcli.tags "
            f'WHERE tag="{tag}" GROUP BY keyword '
            ") tm "
            "ON t.keyword=tm.keyword AND t.timestamp=tm.max_timestamp "
            f'WHERE tag="{tag}"; '
        )
        if not success:
            list_of_keywords = []
            break

        list_of_keywords_ = [thing[0] for thing in output if thing[1] == b"\x01"]

        if return_timestamp:
            for thing in output:
                if thing[1] == b"\x01":
                    timestamp[thing[0]] = thing[2]

        list_of_keywords = (
            list_of_keywords_
            if list_of_keywords is None
            else [
                keyword for keyword in list_of_keywords if keyword in list_of_keywords_
            ]
        )

    table.disconnect()

    list_of_keywords = [] if list_of_keywords is None else sorted(list_of_keywords)

    if after:
        list_of_keywords = [keyword for keyword in list_of_keywords if keyword >= after]

    if before:
        list_of_keywords = [
            keyword for keyword in list_of_keywords if keyword <= before
        ]

    excluded_keywords = reduce(
        lambda x, y: x + y,
        [
            search(
                tag,
                after=after,
                before=before,
                count=-1,
                host=host,
            )
            for tag in excluded_tags
        ],
        [],
    )

    list_of_keywords = [
        keyword for keyword in list_of_keywords if keyword not in excluded_keywords
    ]

    if shuffle:
        random.shuffle(list_of_keywords)
    else:
        list_of_keywords = list_of_keywords[::-1]

    p = re.compile("([0-9]{13}|(0|1)[0-9,a-z]{15}|i-[0-9,a-z]{17})")
    if host == 1:
        list_of_keywords = [keyword for keyword in list_of_keywords if p.match(keyword)]
    if host == 0:
        list_of_keywords = [
            keyword for keyword in list_of_keywords if not p.match(keyword)
        ]

    list_of_keywords = list_of_keywords[offset:]

    list_of_keywords = (
        list_of_keywords[:count]
        if count > 0
        else [] if count != -1 else list_of_keywords
    )

    return (list_of_keywords, timestamp) if return_timestamp else list_of_keywords


def set_(keyword, tags):
    """set tags for keyword.

    Args:
        keyword (str): keyword.
        tags (List(str)): list of strings or string.

    Returns:
        bool: success.
    """
    table = Table(name="tags")

    if isinstance(tags, list):
        tags = ",".join(tags)
    if isinstance(tags, str):
        tags = Options(tags)

    if not table.connect():
        return False

    tags = {tag.strip(): value for tag, value in tags.items()}

    success = True
    for tag in tags:
        if not table.insert(
            "keyword,tag,value".split(","),
            [keyword, tag, 1 if tags[tag] else 0],
        ):
            success = False
        else:
            if tags[tag]:
                logger.info(f"{keyword} += #{tag}.")
            else:
                logger.info(f"{keyword} -= #{tag}.")

    if not table.disconnect():
        return False

    return success



from typing import Any
from .enums import MetadataSourceType
from blue_objects import file


def get(
    key,
    default="",
    source=".",
    source_type: MetadataSourceType = MetadataSourceType.FILENAME,
    filename: str = "metadata.yaml",
    dict_keys: bool = False,
    dict_values: bool = False,
) -> Any:
    success, metadata = file.load_yaml(
        source_type.filename(source, filename),
        ignore_error=True,
    )
    if not success:
        return default

    try:
        output = metadata
        for key_ in [key_ for key_ in key.split(".") if key_]:
            output = output.get(key_, default)

        if dict_keys:
            output = list(output.keys())
        elif dict_values:
            output = list(output.values())

        return output
    except Exception as e:
        return type(e).__name__


def get_from_file(
    filename: str,
    key,
    default="",
    **kwargs,
) -> Any:
    return get(
        key=key,
        default=default,
        source=filename,
        source_type=MetadataSourceType.FILENAME,
        **kwargs,
    )


def get_from_object(
    object_name: str,
    key,
    default="",
    **kwargs,
) -> Any:
    return get(
        key=key,
        default=default,
        source=object_name,
        source_type=MetadataSourceType.OBJECT,
        **kwargs,
    )


def get_from_path(
    path: str,
    key: str,
    default="",
    **kwargs,
) -> Any:
    return get(
        key=key,
        default=default,
        source=path,
        source_type=MetadataSourceType.PATH,
        **kwargs,
    )



import os
from enum import Enum, auto

from blue_objects import objects
from blue_objects.env import abcli_object_path


class MetadataSourceType(Enum):
    FILENAME = auto()
    OBJECT = auto()
    PATH = auto()

    def filename(self, source: str, filename="metadata.yaml"):
        if self == MetadataSourceType.FILENAME:
            return source

        if self == MetadataSourceType.OBJECT:
            return os.path.join(
                objects.object_path(source),
                filename,
            )

        if self == MetadataSourceType.PATH:
            return os.path.join(
                abcli_object_path if source == "." else source,
                filename,
            )

        return f"unknown-source-type:{self.name.lower()}"



import argparse
from . import NAME, pack, KEYWORDS

list_of_tasks = "pack|unpack"

parser = argparse.ArgumentParser(NAME)
parser.add_argument(
    "task",
    type=str,
    default="",
    help=list_of_tasks,
)
parser.add_argument(
    "--keyword",
    type=str,
    default="",
)
parser.add_argument(
    "--default",
    type=str,
    default="",
)
args = parser.parse_args()

success = args.task in list_of_tasks.split("|")
if args.task == "pack":
    print(
        (lambda output: args.default if output in ",-".split(",") else output)(
            pack(args.keyword)
        )
    )
elif args.task == "unpack":
    print(
        (lambda output: args.default if output in ",-".split(",") else output)(
            KEYWORDS.get(args.keyword, args.keyword)
        )
    )
else:
    print(f"-{NAME}: {args.task}: command not found")

if not success:
    print(f"-{NAME}: {args.task}: failed")



import os
import os.path
import pymysql
from abcli import env

from blue_options.logger import crash_report

from abcli.table import NAME
from abcli.logger import logger


class Table:
    def __init__(self, name):
        self.name = name

        self.db = env.ABCLI_AWS_RDS_DB
        self.host = env.ABCLI_AWS_RDS_HOST
        self.password = env.ABCLI_AWS_RDS_PASSWORD
        self.port = int(env.ABCLI_AWS_RDS_PORT)
        self.user = env.ABCLI_AWS_RDS_USER

        self.connection = None

    def connect(self, create_command=""):
        """connect to self.

        Args:
            create_command (str, optional): create command. Defaults to "".

        Returns:
            bool: success.
        """
        if self.connection is not None:
            self.disconnect()

        try:
            self.connection = pymysql.connect(
                host=self.host,
                user=self.user,
                port=self.port,
                password=self.password,
                database=self.db,
            )
        except:
            crash_report(f"-{NAME}: connect: failed on host: {self.host}.")
            return False

        return True if not create_command else self.create(create_command)

    @staticmethod
    def Create(table_name, create_command):
        """create table_name.

        Args:
            table_name (str): table name.
            create_command (List[str]): create command.

        Returns:
            bool: success
        """
        table = Table(name=table_name)

        return table.disconnect() if table.connect(create_command) else False

    def create(self, create_command):
        """create self.

        Args:
            create_command (List[str]): create command.

        Returns:
            bool: success.
        """
        return self.execute(
            "CREATE TABLE IF NOT EXISTS {} ({})".format(
                self.name,
                ",".join(
                    [
                        "id INT(24) NOT NULL AUTO_INCREMENT",
                        "timestamp TIMESTAMP DEFAULT CURRENT_TIMESTAMP",
                    ]
                    + create_command
                    + ["PRIMARY KEY (`id`)", "INDEX `index_timestamp` (`timestamp`)"]
                ),
            ),
            commit=True,
        )

    def disconnect(self):
        """disconnect.

        Returns:
            bool: success.
        """
        if self.connection is None:
            return True

        success = True
        try:
            self.connection.close()
        except:
            crash_report(f"-{NAME}: disconnect: failed.")
            success = False

        self.connection = None
        return success

    def drop(self):
        """drop self.

        Returns:
            bool: success.
        """
        return self.execute(f"DROP table {self.name};")

    def execute(self, sql, commit=False, returns_output=True):
        """execute sql command.

        Args:
            sql (str): sql command.
            commit (bool, optional): commit changes. Defaults to False.
            returns_output (bool, optional): return output. Defaults to True.

        Returns:
            bool: success.
            Any: output, optional.
        """
        output = []
        success = False
        try:
            with self.connection.cursor() as cursor:
                if isinstance(sql, tuple):
                    cursor.execute(sql[0], sql[1])
                else:
                    cursor.execute(sql)

                if returns_output:
                    output = cursor.fetchall()

                if commit:
                    # connection is not autocommit by default. So you must commit to save
                    # your changes.
                    self.connection.commit()

            success = True
        except:
            crash_report(f"-{NAME}: execute({sql}): failed.")

        return (success, output) if returns_output else success

    def insert(self, columns, values):
        """insert data in self.

        Args:
            columns (List[str]): list of columns to insert.
            values (List[Any]): list of values to insert.

        Returns:
            bool: success
        """
        return self.execute(
            (
                f"INSERT INTO {self.name}"
                + " ("
                + ", ".join(columns)
                + ") VALUES ("
                + ", ".join(len(columns) * ["%s"])
                + ")",
                values,
            ),
            commit=True,
            returns_output=True,
        )



import argparse
from . import *
from abcli.logger import logger
from blueness.argparse.generic import sys_exit


parser = argparse.ArgumentParser(NAME)
parser.add_argument(
    "task",
    type=str,
    default="terraform",
    help="poster,terraform",
)
parser.add_argument(
    "--filename",
    type=str,
    default="background.jpg",
)
parser.add_argument(
    "--is_headless",
    type=str,
    default="false",
    help="true/false",
)
parser.add_argument(
    "--target",
    type=str,
    default="none",
    help="lxde/mac/rpi/ubuntu",
)
parser.add_argument(
    "--user",
    type=str,
    default="user",
)
args, _ = parser.parse_known_args()

success = False
if args.task == "poster":
    success = poster(args.filename)
elif args.task == "terraform":
    logger.info("terraforming {}".format(args.target))

    if args.target == "lxde":
        success = lxde(args.user)
    elif args.target == "mac":
        success = mac(args.user)
    elif args.target == "rpi":
        success = rpi(
            args.user,
            is_headless=args.is_headless == "true",
        )
    elif args.target == "ubuntu":
        success = ubuntu(args.user)
    else:
        logger.error(f"-{NAME}: {args.target}: target not found.")
else:
    success = None

sys_exit(logger, NAME, args.task, success)
